
% Any publication resulting from the use of this m-file shall acknowledge
% it by citing the following paper:

%S. P. Torres and C. A. Castro, Expansion planning for smart transmission
%grids using ac model and shunt compensation, IET Generation,
%Transmission & Distribution, vol. 8, no. 5, pp. 966-975, 2014.

function [xopt,fxopt,fevalcount,FOPT,RunStats] = CE_EPSO_opt(MaxIt,Max_FES,SS,dim,Xmin,Xmax,GM,nexp,fhd,sistema,varargin)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Initialization of parameters needed for CE-EPSO
clc;
rand('state',sum(100*clock)); % Different random generator seeds per execution


Acc = 1e-3; % Desired accuracy
RunStats = [ ]; % Matrix of average statistics

xopt = zeros(dim, nexp);
FOPT = zeros(MaxIt, nexp);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%===================================
% Iterative loop
%===================================
for T = 1:nexp
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Other initializations
    success = 0; % Exit flag
    iter = 0; % Iteration counter
    fevalcount = 0; % Counter of objective function evaluations
    STOP = 0; % Stop flag
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% CE Initializations
    D = dim;
    % Individuals' lower bounds.
    x_min = Xmin(:,1)';
    % Individuals' upper bounds.
    x_max = Xmax(:,1)';
    n_samples = SS;
    Nelite = 20;
    sigma = 0.9;
    indexRo = n_samples - Nelite;
    pos = zeros( n_samples, D );
    gammaIter = zeros( 1e6, 1 );
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    max_iter_ce = 10;
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% EPSO Initializations
    global epso_par;
    epso_par.pop_size = SS * 0.25;
    epso_par.mutation_rate = 0.4;
    epso_par.communication_probability = 0.8;
    % Weights matrix
    % 1 - inertia
    % 2 - memory
    % 3 - cooperation
    % 4 - perturbation
    weights = rand( epso_par.pop_size, 4 ) * 1.5;
    epso_par.init = 1;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Update global best
    gbestval = 1e30;
    gbest = zeros( 1, D );
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Initialize CE METHOD
    n_trials = zeros( 1, D );
    meanPar = zeros( 1, D );
    for i = 1 : D
        meanPar( 1, i ) = 0.5;
        n_trials( 1, i ) = x_max( 1, i ) - x_min( 1, i );
    end
    iter_ce = 0;
    memMeanPar( iter_ce + 1, : ) = meanPar;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Main loop
    while (STOP == 0)
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Run CE METHOD
        while iter_ce < max_iter_ce
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Update global iteration
            iter = iter + 1;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %% Generate population
            for i = 1 : D
                tmp_stat = binornd( n_trials( i ), meanPar( i ), 1, n_samples );
                tmp_stat = tmp_stat + x_min( 1, i );
                tmp_stat( tmp_stat < x_min( 1, i ) ) = x_min( 1, i );
                tmp_stat( tmp_stat > x_max( 1, i ) ) = x_max( 1, i );
                pos( :, i ) = tmp_stat;
            end
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Evaluate population
            fit = zeros( n_samples, 1 );
            costo_lineas = zeros( n_samples,1 );
            penalizacion = zeros( n_samples, 1 );
            pos_eval = pos';
             parfor i = 1: n_samples
            % for i = 1: n_samples
                [costo_lineas(i), penalizacion(i)] = feval( fhd, pos_eval( :, i ), sistema );
                fit(i) = costo_lineas(i) + penalizacion(i);
                fevalcount = fevalcount + 1;
            end
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % UPDATE GLOBAL BEST
            [ tmpgbestval, gbestid ] = min( fit );
            if tmpgbestval < gbestval
                gbestval = tmpgbestval;
                gbest = pos( gbestid, : );
            end
            fxopt = gbestval;
            xopt(:,T) = gbest';
            FOPT(iter,T) = fxopt;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Performs CE correction
            
            % Sorts states according to fitness
            [ Saux, Iaux ] = sort( fit, 'descend' );
            
            H = zeros( 1, n_samples );
            for i = 1 : n_samples
                if i <= indexRo
                    H( Iaux( i ) ) = 0;
                else
                    H( Iaux( i ) ) = 1;
                end
            end
            
            % Computes gammaIter
            gamma = Saux( indexRo );
            gammaIter( iter_ce + 1, 1 ) = gamma;
            
            % Applies CE correction
            den = sum( H );
            for i = 1 : D
                tmp = pos( :, i ) - x_min( 1, i );
                num = H * ( tmp );
                tmpMean = 0;
                if den > 0
                    tmpMean = ( num / den ) * ( 1 / n_trials( 1, i ) );
                end
                meanPar( i ) = tmpMean * sigma + meanPar( i ) * ( 1 - sigma );
                
                % Update sampling distribution parameters
                meanPar( i ) = round( meanPar( i ), 2 );
                if meanPar( i ) < 0.01
                    meanPar( i ) = 0.01;
                elseif meanPar( i ) > 0.99
                    meanPar( i ) = 0.99;
                end
            end
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Update CE iteration
            iter_ce = iter_ce + 1;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            memMeanPar( iter_ce + 1, : ) = meanPar;
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Run EPSO
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %  Initialize EPSO -> just once
        if epso_par.init == 1
                        
            epso_par.Xmin = x_min - 0.499;
            epso_par.Xmax = x_max + 0.499;
            
            Vmin = (-epso_par.Xmax + epso_par.Xmin)*1;
            Vmax = -Vmin;
            vel = zeros( epso_par.pop_size, D );
            tmp_pos = zeros( epso_par.pop_size, D );
            tmp_fit = zeros( epso_par.pop_size, 1 );
            for i = 1 : epso_par.pop_size
                tmp_pos( i, : ) = pos( Iaux( n_samples - i + 1 ), : );
                tmp_fit( i ) = fit( Iaux( n_samples - i + 1 ) );
            end
            pos = tmp_pos;
            fit = tmp_fit;
            
            for i = 1 : epso_par.pop_size
                 vel( i, : ) = ( Vmin + ( Vmax - Vmin ) .* rand( 1, D ) );
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % UPDATE individual best
            % Individual best position ever of the particles of current population
            myBestPos = pos;
            % Fitness of the individual best position ever of the particles of current population
            myBestPosFit = fit;
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            epso_par.init = 0;
            
            gammaIter = gammaIter( 1:iter_ce, 1 );
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Update global iteration
        iter = iter + 1;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % COPY CURRENT population
        copy_pos = pos;
        copy_vel = vel;
        copy_weights = weights;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Apply EPSO movement rule
        for i = 1 : epso_par.pop_size
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Mutate weights of the particles of the COPIED population
            copy_weights( i, :, : ) = MUTATE_WEIGHTS( weights( i, : ), epso_par.mutation_rate );
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Compute new velocity for the particles of the COPIED population
            copy_vel( i, : ) = EPSO_COMPUTE_NEW_VEL( D, copy_pos( i, : ), myBestPos( i, : ), gbest, ...
                copy_vel( i, : ), Vmin, Vmax, copy_weights( i, : ), epso_par.communication_probability );
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            if rand() < 0.5
                % Compute new position for the particles of the COPIED population
                [ copy_pos( i, : ), copy_vel( i, : ) ] = COMPUTE_NEW_POS( copy_pos( i, : ), copy_vel( i, : ), epso_par.Xmin, epso_par.Xmax );
                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            else
                % Compute new position for the particles of the COPIED population
                [ copy_pos( i, : ), copy_vel( i, : ) ] = COMPUTE_NEW_POS_TEP( copy_pos( i, : ), copy_vel( i, : ), epso_par.Xmin, epso_par.Xmax );
                %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            end
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % Compute new velocity for the particles of the CURRENT population
            vel( i, : ) = EPSO_COMPUTE_NEW_VEL( D, pos( i, : ), myBestPos( i, : ), gbest, ...
                vel( i, : ), Vmin, Vmax, weights( i, : ), epso_par.communication_probability );
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            if rand() < 0.5
                % Compute new position for the particles of the CURRENT population
                [ pos( i, : ), vel( i, : ) ] = COMPUTE_NEW_POS( pos( i, : ), vel( i, : ), epso_par.Xmin, epso_par.Xmax );
            else
                % Compute new position for the particles of the CURRENT population
                [ pos( i, : ), vel( i, : ) ] = COMPUTE_NEW_POS_TEP( pos( i, : ), vel( i, : ), epso_par.Xmin, epso_par.Xmax );
            end
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Enforce search space limits of the COPIED population
        [ copy_pos, copy_vel ] = ENFORCE_POS_LIMITS( D, epso_par.pop_size, copy_pos, epso_par.Xmin, epso_par.Xmax, copy_vel, Vmin, Vmax );
        copy_pos = round( copy_pos );
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Enforce search space limits of the CURRENT population
        [ pos, vel ] = ENFORCE_POS_LIMITS( D, epso_par.pop_size, pos, epso_par.Xmin, epso_par.Xmax, vel, Vmin, Vmax );
        pos = round( pos );
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Evaluate the COPIED population
        copy_fit = zeros( epso_par.pop_size, 1 );
        copy_costo_lineas = zeros( epso_par.pop_size, 1 );
        copy_penalizacion = zeros( epso_par.pop_size, 1 );
        copy_pos_eval = copy_pos';
         parfor i = 1 : epso_par.pop_size
        % for i = 1: epso_par.pop_size
            [copy_costo_lineas(i), copy_penalizacion(i)] = feval( fhd, copy_pos_eval( :, i ), sistema );
            copy_fit(i) = copy_costo_lineas(i) + copy_penalizacion(i);
            fevalcount = fevalcount + 1;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Evaluate the CURRENT population
        fit = zeros( epso_par.pop_size, 1 );
        costo_lineas = zeros( epso_par.pop_size, 1 );
        penalizacion = zeros( epso_par.pop_size, 1 );
        pos_eval = pos';
        parfor i = 1 : epso_par.pop_size
      % for i = 1: epso_par.pop_size
            [costo_lineas(i), penalizacion(i)] = feval( fhd, pos_eval( :, i ), sistema );
            fit(i) = costo_lineas(i) + penalizacion(i);
            fevalcount = fevalcount + 1;
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % CREATE NEW population to replace CURRENT population
        selParNewSwarm = ( copy_fit < fit );
        for i = 1 : epso_par.pop_size
            if selParNewSwarm( i )
                fit( i ) = copy_fit( i );
                pos( i, : ) = copy_pos( i, : );
                vel( i, : ) = copy_vel( i, : );
                weights( i, : ) = copy_weights( i, : );
            end
            if fit( i ) < myBestPosFit( i )
                myBestPos( i, : ) = pos( i, : );
                myBestPosFit( i ) = fit( i );
            end
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % UPDATE GLOBAL BEST
        [ tmpgbestval, gbestid ] = min( fit );
        if tmpgbestval < gbestval
            gbestval = tmpgbestval;
            gbest = pos( gbestid, : );
        end
        fxopt = gbestval;
        xopt(:,T) = round( gbest' );
        FOPT(iter,T) = fxopt;
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % CHECK STOP CRITERION
        if ((fevalcount >= Max_FES)||(fxopt <= GM+Acc))
            STOP = 1;
            if (fxopt <= GM+Acc)
                success = 1;
            end
        end
       if ((iter >= MaxIt)||(fxopt <= GM+Acc))
            STOP = 1;
            if (fxopt <= GM+Acc)
                success = 1;
            end
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
    end
    
    %% UPDATE VECTOR OF STATISTICS
    RunStats(T,:) = [success iter fevalcount];
    
end